#include "K575.h"

char Buf[200];	

#pragma export on

uint8			ValueOfPortC				= 0,
				ValueOfPortA				= 0;
bool			ADCIsConverting				= false;

void 			*Address_Space				= 0,
				*ISA_Memory					= 0;
bool			DriverIsOpen				= false;
volatile
bool			AcquisitationInProgress 	= false;
static
const 
char 			*TheDeviceName[] 			= { "K575", NULL };
uint16			*LeftRawDataBuffer,
				*RightRawDataBuffer;
int32			ValueOfXDAC					= 0,
				ValueOfYDAC					= 0,
				BiasDARange,
				CurrentDARange,
				MaxValDA					= 4096,
				KEITHLY_Offset 				= 0x9ff80,
				KEITHLY_Page_Size 			= 0x80,
				BiasLSBRegister,
				BiasMSBRegister,
				CurrentLSBRegister,
				CurrentMSBRegister,
				KernelAreaCounter 			= 0;

thread_id		AcquisitationThreadID		= 0;
physical_entry	Phys_Table[1];
static
device_hooks 	devices[] 					= {	{
													K575_open,
													K575_close,
													K575_free_cookie,
													K575_control,
													K575_read,
													K575_write
												},
 												0
 											  };

// The Internal Timing is allocated as a kernal area via an ioctl-call (STM_CREATE_AREA).
// That is because both BeaSTM and the driver must be able to access the area
// at runtime.
STMImageWindow		*CallingWindow;
InternalScanning	*Scanning;
InternalTiming		*Timing;
InternalImageParams	*ImageParams;
float				LeftBias,
					RightBias,
					LeftCurrent,
					RightCurrent;
//
// Functions to implement:
//

void
MoveTipToOrigo( )
{
	int32	x,
			y;

	if( ValueOfXDAC > 0)
	{
		SetXDACStepDirectionTo( DOWN );
		for( x = ValueOfXDAC; x > 0; x-- )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneXPixel(	1, Timing->DADelay );
			ValueOfXDAC--;
		}
	}
	else
	{
		SetXDACStepDirectionTo( UP );
		for( x = ValueOfXDAC; x < 0; x++ )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneXPixel(	1, Timing->DADelay );
			ValueOfXDAC++;
		}
	}
	
	if( ValueOfYDAC > 0)
	{
		SetYDACStepDirectionTo( DOWN );
		for( y = ValueOfYDAC; y > 0; y-- )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneYPixel(	1, Timing->DADelay );
			ValueOfYDAC--;
		}
	}
	else
	{
		SetYDACStepDirectionTo( UP );
		for( y = ValueOfYDAC; y < 0; y++ )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneYPixel(	1, Timing->DADelay );
			ValueOfXDAC++;
		}
	}

//	ValueOfXDAC		= 0;
//	ValueOfYDAC		= 0;
}

void
MoveTipToDAPoint( uint32 argXPosition, uint32 argYPosition )
{
	int32	x,
			y;

	dprintf("MoveTipToDAPoint: Start: Tip(%d, %d)\n", ValueOfXDAC,ValueOfYDAC);
	dprintf("MoveTipToDAPoint: Point to move: Tip(%d, %d)\n", argXPosition,argYPosition);

	if( ValueOfXDAC > argXPosition)
	{
		SetXDACStepDirectionTo( DOWN );
		for( x = ValueOfXDAC; x > argXPosition; x-- )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneXPixel(	1, Timing->DADelay );
			ValueOfXDAC--;
		}
	}
	else
	{
		SetXDACStepDirectionTo( UP );
		for( x = ValueOfXDAC; x < argXPosition; x++ )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneXPixel(	1, Timing->DADelay );
			ValueOfXDAC++;
		}
	}

	dprintf("MoveTipToDAPoint: Moved X: Tip(%d, %d)\n", ValueOfXDAC, ValueOfYDAC);
	
	if( ValueOfYDAC > argYPosition)
	{
		SetYDACStepDirectionTo( DOWN );
		for( y = ValueOfYDAC; y > argYPosition; y-- )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneYPixel(	1, Timing->DADelay );
			ValueOfYDAC--;
		}
	}
	else
	{
		SetYDACStepDirectionTo( UP );
		for( y = ValueOfYDAC; y < argYPosition; y++ )
		{
			microsecond_delay(	Timing->FlyBackDelay );
			StepOneYPixel(	1, Timing->DADelay );
			ValueOfYDAC++;
		}
	}

	dprintf("MoveTipToDAPoint: Moved X & Y: Tip(%d, %d)\n", ValueOfXDAC, ValueOfYDAC);
}


int16
TakeSamplesFromADC( int32		argNumberOfSamplesToTake )
{
	int32		NumberOfSamplesTaken	= 0,
				Accumulator				= 0;

	// Start conversion:
	KWrite( 255,  CMDD1 );
	// This is taken from P. Mårtensson and F. Ownmans program and I don't
	// Dummy read to clear first EOT/EOC *)
	KRead( CMDA1 );

	// Wait until bit 7 of CMDD1 is cleared:
	while( KRead( CMDD1 ) >127 )
		;
		
	while( NumberOfSamplesTaken < argNumberOfSamplesToTake )
	{
		// Start conversion:
		KWrite( 255, CMDD1 );

		// This is taken from P. Mårtensson and F. Ownmans program and I don't
		// Clear EOT/EOC:
       
       Accumulator += ( (uint16)KRead( CMDB1 ) << 8 ) | KRead( CMDA1 );       
		
		NumberOfSamplesTaken++;

		// Wait until done, i.e. until bit #7 of CMDD1 is no longer set
		while( KRead( CMDD1 ) > 127 )
			;
	}
	// Increment data: 256*MSB + LSB and calculate output:
	return Accumulator/NumberOfSamplesTaken;
}

status_t
ResetAD()
{
	int64	StartTime = system_time();
	bool	TimeOut = false;

	/* Set read status: bit #4 is not set */
	/* Slot #1: of bits #0 - #3 only bit #0 is set
		Set range bipolar: bit #5 is set
		Set global gain 1: bits #6 and #7 both not set */
	KWrite( 33, CMDB1 );

	/* Reset and recal the A/D converter */
	KWrite( 255, CMDC1 );

	/* Wait until done, i.e. until bit #7 of CMDA1 is no longer set */
	while( KRead( CMDA1 ) > 127 && ! TimeOut)
	{
		if( system_time() - StartTime > 1000000 )
			TimeOut = true;
	}
	
	if( TimeOut )
	{
		D("Time out!");
		return B_ERROR;	
	}
	
	/* Set read data: bit #4 is set */
	/* Slot #1: of bits #0 - #3 only bit #0 is set
		Set range bipolar: bit #5 is set
		Set global gain 1: bits #6 and #7 both not set */
	KWrite( 49, CMDB1 );
	return B_OK;

	ValueOfXDAC 	= 0;
	ValueOfYDAC 	= 0;
}

status_t
InitExternalDAC()
{
	// Digital I/O:
	// Port A and C are directed out, B and D in.
	// Port A: bit 3, B: bit 4, C: bit 5, D: bit 6.
	// => bit(6,5,4,3) = (0,1,0,1) => CMDA5 = 40
	// Set the configuration bit (7) => CMDA5 = 40+128 = 168
	KWrite( 168, CMDA5 );
	
	// X and Y for (x,y) positioning, given by port C:
	ValueOfPortC = 0;
	// Select port C by setting the bits CMDA5 to %00000010 = 2:
	KWrite( 2, CMDA5 );
	KWrite( ValueOfPortC , CMDB5 );

	// Reset X:
	// Set bit 4 of Port C low and high again:	
	ValueOfPortC &= CLEAR_BIT( 4 );
	KWrite( ValueOfPortC, CMDB5 );
	ValueOfPortC |= SET_BIT( 4 );
	KWrite( ValueOfPortC, CMDB5 );

	//Reset Y:
	// Set bit 5 of Port C low and high again:	
	ValueOfPortC &= CLEAR_BIT( 5 );
	KWrite( ValueOfPortC, CMDB5 );
	ValueOfPortC |= SET_BIT( 5 );
	KWrite( ValueOfPortC, CMDB5 );

	// x and y direction: up
	SetXDACStepDirectionTo( UP );
	SetYDACStepDirectionTo( DOWN );
	
	// Track and hold: port C
	// Set bit 7 of port C low
	// Bit 7 cleared
	ValueOfPortC &= CLEAR_BIT( 7 );
	KWrite( ValueOfPortC, CMDB5 );
 	
	// Hardware interface: port A
	ValueOfPortA = 0;
	// Select port A by setting the bits CMDA5 to %00000000 = 0:
	KWrite( 0, CMDA5 );
	KWrite( ValueOfPortA , CMDB5 );
	
	return B_OK;
}

status_t
ConfigADCAndDACFromParameters( InternalADDA *Params )
{
	uint8	ValueToWriteToCMDB1 = 33+16,
			ValueToWriteToCMDA1 = Params->ADCChannel;

	// Local Gain:
	
	switch(	Params->ADCLocalAmplification )
	{
		case 1:
			ValueToWriteToCMDA1 |= 0;
			break;

		case 10:
			ValueToWriteToCMDA1 |= SET_BIT( 5 );
			break;

		default:
			D("Invalid Local Gain!!\n");
	}

	// Filter:
	
	switch(	Params->ADCFilter )
	{
		case 1:
			ValueToWriteToCMDA1 |= SET_BIT( 7 );
			break;

		case 100:
			ValueToWriteToCMDA1 |= 0;
			break;

		default:
			D("Invalid Filter!!\n");
	}

	KWrite( ValueToWriteToCMDA1, CMDA1 );

	SetDARange( Params->BiasChannel, Params->DACRangeOfBiasChannel  );
	SetDARange( Params->CurrentChannel, Params->DACRangeOfCurrentChannel  );

	// The global amplification:

	switch(	Params->ADCGlobalAmplification )
	{
		case 1:
			ValueToWriteToCMDB1 |= 0;
			break;

		case 2:
			ValueToWriteToCMDB1 |= SET_BIT( 6 );
			break;

		case 5:
			ValueToWriteToCMDB1 |= SET_BIT( 7 );
			break;

		case 10:
			ValueToWriteToCMDB1 |= SET_BIT( 7 ) | SET_BIT( 6 );
			break;

		default:
			D("Invalid Global Gain!!\n");
	}

	BiasLSBRegister 	= Params->BiasChannel*2;
	BiasMSBRegister 	= Params->BiasChannel*2+1;
	CurrentLSBRegister	= Params->CurrentChannel*2;
	CurrentMSBRegister 	= Params->CurrentChannel*2+1;

	KWrite( ValueToWriteToCMDB1, CMDB1 );

	return B_OK;
}

status_t
InitADCAndDAC()
{
	// Check if connection with K575 and see if AMM2 module is installed
	if( KRead( BOX_ID ) != 0x72 )
	{
		D("Could not find the K575 hardware!\n");
		dprintf("Error code: %d\n", STM_CANNOT_INIT_HARDWARE );
		return STM_CANNOT_INIT_HARDWARE;
	}

   // Try to initialize A/D converter:
   if( ResetAD() != B_OK )
	{
		D("Could not reset the A/D!\n");
		return STM_CANT_RESET_THE_ADS;
	}

	// Check if AOM5 is installed 
	// K575 has two physical slots, we check for #3.
	// That means we code for it as 12.
	// Set read data: Bit 4 is set.
	// Set the range as unipolar: Bit 4 cleared.
	// Set global gain 1: Bit 6 and 7 cleared.
	KWrite( 12 | SET_BIT(4), CMDB1 );

	snooze(5000);
	
	uint8	IDofAOM5 = ( TakeSamplesFromADC(50) / 100 - 1 ) & 255;

	dprintf("IDofAOM5 = %d\n", IDofAOM5 );
	
	if( IDofAOM5 < 57 || IDofAOM5 > 59 )
	{
		D("AOM5 not installed in slot 3.\n");
		return STM_AOM5_NOT_INSTALLED;
	}

	// Initialize:
	// Digital I/O and external D/A converter
	InitExternalDAC();
	
	// ADC:
//	ConfigADCAndDACFromParameters( );

	// D/A converters range and initial values:
	SetDARange( 0, 1 );
	SetDAValue( 0, 0 );
	SetDARange( 1, 1 );
	SetDAValue( 1, 0 );
	SetDARange( 2, 1 );
	SetDAValue( 2, 0 );
	SetDARange( 3, 1 );
	SetDAValue( 3, 0 );
	SetDARange( 4, 1 );
	SetDAValue( 4, 0 );
	SetDARange( 5, 1 );
	SetDAValue( 5, 0 );
	
	return B_OK;
}

status_t
SetDARange( int32	argChannel, int32 argRange )
{
	int32	RangeNumber[] = { 0, 3, 2, 0, 0, 1, 0, 0, 0, 0, 0 };

	if( argChannel < 0 || argChannel > 5 )
		return B_ERROR;
	
	if( argRange != 10 && argRange != 5 && argRange != 2 && argRange != 1 )
		return B_ERROR;
		
	switch( argChannel )
	{
		case 0: /* Channel 0: Register 15 */
			KWrite( 15, CMDA4 );
			KWrite( RangeNumber[ argRange ], CMDB4 );
			break;

		case 1: /* Channel 1: Register 14 */
			KWrite( 14, CMDA4 );
			KWrite( RangeNumber[ argRange ], CMDB4 );
			break;

		case 2: /* Channel 0 of AOM5: Register 15 */
			KWrite( 15, CMDA9 );
			KWrite( RangeNumber[ argRange ], CMDB9 );
			break;

		case 3: /* Channel 1 of AOM5: Register 14 */
			KWrite( 14, CMDA9 );
			KWrite( RangeNumber[ argRange ], CMDB9 );
			break;

		case 4: /* Channel 2 of AOM5: Register 13 */
			KWrite( 13, CMDA9 );
			KWrite( RangeNumber[ argRange ], CMDB9 );
			break;

		case 5: /* Channel 3 of AOM5: Register 12 */
			KWrite( 12, CMDA9 );
			KWrite( RangeNumber[ argRange ], CMDB9 );
			break;		
	}
	return B_OK;
}

status_t
SetDAValue(int32	argChannel, int32 argValue )
{
	uint8	Hi8Bits,
			Lo8Bits;
	int32	RangeNumber[] = { 0,3,2,0,0,1,0,0,0,0,0};

	if( argChannel < 0 || argChannel > 5 )
		return B_ERROR;

	if( argValue < 0 )
	{
		argValue = -argValue;
		argValue |= SET_BIT( 15 );
	}
	Lo8Bits = argValue;
	Hi8Bits	= argValue >> 8;

	switch( argChannel )
	{
		case 0: // Channel 0. Reg. 0
			KWrite( 0, CMDA4 );
			KWrite( Lo8Bits, CMDB4 );
			KWrite( 1, CMDA4 );
			KWrite( Hi8Bits, CMDB4 );
			break;
		case 1: // Channel 1. Reg. 1
			KWrite( 2, CMDA4 );
			KWrite( Lo8Bits, CMDB4 );
			KWrite( 3, CMDA4 );
			KWrite( Hi8Bits, CMDB4 );
			break;
		case 2: // Channel 0 of AOM5. Reg. 0
			KWrite( 0, CMDA9 );
			KWrite( Lo8Bits, CMDB9 );
			KWrite( 1, CMDA9 );
			KWrite( Hi8Bits, CMDB9 );
			break;
		case 3: // Channel 1 of AOM5. Reg. 0
			KWrite( 2, CMDA9 );
			KWrite( Lo8Bits, CMDB9 );
			KWrite( 3, CMDA9 );
			KWrite( Hi8Bits, CMDB9 );
			break;
		case 4: // Channel 2 of AOM5. Reg. 0
			KWrite( 4, CMDA9 );
			KWrite( Lo8Bits, CMDB9 );
			KWrite( 5, CMDA9 );
			KWrite( Hi8Bits, CMDB9 );
			break;
		case 5: // Channel 3 of AOM5. Reg. 0
			KWrite( 6, CMDA9 );
			KWrite( Lo8Bits, CMDB9 );
			KWrite( 7, CMDA9 );
			KWrite( Hi8Bits, CMDB9 );
			break;
	}
	KWrite( 255, STROBE );
	
	return B_OK;
}

inline 
uint16
CalcNewIntegerBias( float argBias )
{
	int16 NewBias;
	
	NewBias = MaxValDA*argBias/BiasDARange;

	// Bit 15 is a sign bit in the DAC. Set it if
	// the value is negative.
	if( NewBias < 0 )
	{
		// Make sure we have a positive value:
		NewBias = -NewBias;
		// Set the 15:th bit:
		NewBias |= 1<<15;		
	}
	
	return NewBias;	
}

inline 
uint16
CalcNewIntegerCurrent( float argCurrent )
{
	int16 NewCurrent;
	
	NewCurrent = MaxValDA*argCurrent/CurrentDARange;

	// The current must always be negative, so if it's
	// not, set the negative sign. (DAC:s 15:th bit is the sign.)
	if( NewCurrent < 0 )
	{
		NewCurrent = -NewCurrent;
		NewCurrent |= 1<<15;
	}
	else
		NewCurrent |= 1<<15;

	return NewCurrent;	
}

inline
void
SetBiasAndCurrent( float argBias, float argCurrent )
{
	SetIntegerBiasAndCurrent( CalcNewIntegerBias( argBias ), CalcNewIntegerCurrent( argCurrent ));
}

inline
void
SetIntegerBiasAndCurrent( uint16 argBias, uint16 argCurrent )
{
	uint8	BiasLSB		= argBias & 255,
			BiasMSB		= argBias >> 8,
			CurrentLSB	= argCurrent & 255,
			CurrentMSB	= argCurrent >> 8;
	// First, HOLD, by setting bit 7 of port C high:

	// Select port C by setting the bits CMDA5 to %00000010 = 2:
	KWrite( 2, CMDA5 );
	// Set the seventh bit in port C high to HOLD:
	ValueOfPortC	|= SET_BIT( 7 );
	KWrite( ValueOfPortC , CMDB5 );
	
	// Wait the HOLD time:
	microsecond_delay( Timing->HoldTime );
	
	// Send the Bias:

	// Channel 0, Select the Least Significat Byte: register 0 if channel 0 chosen
	KWrite( BiasLSBRegister, CMDA4 );
	// Send the 8 least significat bits:
	KWrite( BiasLSB, CMDB4 );
	// Channel 0, Select the Most Significat Byte: register 1 if channel 0 chosen
	KWrite( BiasMSBRegister, CMDA4 );
	// Send the 4 most significat bits:
	KWrite( BiasMSB, CMDB4 );
	
	// Send the Current:
	
	// Channel 1, Select the Least Significat Byte: register 2 if channel 1 is chosen
	KWrite( CurrentLSBRegister, CMDA4 );
	// Send the 8 least significat bits:
	KWrite( CurrentLSB, CMDB4 );
	// Channel 1, Select the Most Significat Byte: register 3 if channel 1 is chosen
	KWrite( CurrentMSBRegister, CMDA4 );
	// Send the 4 most significat bits:
	KWrite( CurrentMSB, CMDB4 );
	
	// Send Global Strobe:
	
	KWrite( 255, STROBE );
	
	// Wait the HOLD time:
	microsecond_delay( Timing->HoldTime );

	// Turn on the tracking again:
	// Select port C by setting the bits CMDA5 to %00000010=2:
	KWrite( 2, CMDA5 );
	// Clear the seventh bit in port C high to TRACK:
	ValueOfPortC	&= CLEAR_BIT( 7 );
	KWrite( ValueOfPortC , CMDB5 );

	if( ADCIsConverting )
	{
		// This is taken from P. Mårtensson and F. Owmans program and I don't
		// understand why/how one can clear  the EOT/EOC with this dummy read:
		
		KRead( CMDA1 );
		
		// Wait until conversion is finished:
		
		while( KRead( CMDD1 ) > 127 )
			;
	}

}

inline
void
SetXDACStepDirectionTo( bool argDirection )
{
	// Set bit 2 of PortC high if downwards, low if upwards.
	switch( argDirection )
	{
		case UP:
			ValueOfPortC &= CLEAR_BIT( 2 );
			break;
		
		case DOWN:
			ValueOfPortC |= SET_BIT( 2 );
			break;
	
		default:
			D("Programming Error: SetXDACStepDirection: argDirection not UP and not DOWN!\n");
			return;
			break;
	}

	// Select port C by setting the bits CMDA5 to %00000010 = 2:
	KWrite( 2, CMDA5 );
	KWrite( ValueOfPortC , CMDB5 );
}

inline
void
SetYDACStepDirectionTo( bool argDirection )
{
	// Set bit 3 of PortC high if downwards, low if upwards.
	switch( argDirection )
	{
		case UP:
			ValueOfPortC &= CLEAR_BIT( 3 );
			break;
		
		case DOWN:
			ValueOfPortC |= SET_BIT( 3 );
			break;
	
		default:
			D("Programming Error: SetXDACStepDirection: argDirection not UP and not DOWN!\n");
			return;
			break;
	}

	// Select port C by setting the bits CMDA5 to %00000010 = 2:
	KWrite( 2, CMDA5 );
	KWrite( ValueOfPortC , CMDB5 );
}

inline
void
StepOneXPixel(	int32 argNumberOfDASteps, int32 argDelayBetweenSteps )
{
	int32	i;

	microsecond_delay( Timing->PixelDelay );

	for( i=0; i < argNumberOfDASteps; i++ )
	{
		microsecond_delay( argDelayBetweenSteps );

		// Step one step in the X-direction:
		// Select the C-port:
		KWrite( 2, CMDA5 );

		// Set bit 0...
		ValueOfPortC 	|= SET_BIT(0);
		KWrite( ValueOfPortC, CMDB5 );

		// ...and clear it again:
		ValueOfPortC 	&= CLEAR_BIT(0);
		KWrite( ValueOfPortC, CMDB5 );
	}
	
}

inline
void
StepOneYPixel(	int32 argNumberOfDASteps, int32 argDelayBetweenSteps )
{
	int32	i;

	microsecond_delay( Timing->PixelDelay );
	
	for( i=0; i < argNumberOfDASteps; i++ )
	{
		microsecond_delay( argDelayBetweenSteps );

		// Step one step in the Y-direction
		// Select the C-port:
		KWrite( 2, CMDA5 );

		// Set bit 1...
		ValueOfPortC 	|= SET_BIT(1);
		KWrite( ValueOfPortC, CMDB5 );

		// ...and clear it again:
		ValueOfPortC 	&= CLEAR_BIT(1);
		KWrite( ValueOfPortC, CMDB5 );
	}
}


inline 
void
microsecond_delay( int64 x )
{
	if( x==0 )
		return;

	cpu_status stat = disable_interrupts();

	int64 time_to_go = system_time()+x;
	
	while( system_time() < time_to_go )
		;
		
	restore_interrupts( stat );
}

void
KWrite( uint8 value,int command_register )
{
	*((uint8*)ISA_Memory+KEITHLY_Offset+command_register) = value;
}

uint8
KRead( int command_register )
{
	return  *((uint8 *)ISA_Memory+KEITHLY_Offset+command_register);
}

void
DoPreScans( uint32 argNumberOfPrescans )
{
	uint16			x					= 0,
					y					= 0,
					XSize				= Scanning->ScanXSize,
					YSize				= Scanning->ScanYSize,
	// Left/RightBias/Current variables are set in 
	// the ioctl()-call STM_PREPAE_SCAN below.
					LeftBias16Bit		= CalcNewIntegerBias( LeftBias ),
					RightBias16Bit		= CalcNewIntegerBias( RightBias ),
					LeftCurrent16Bit	= CalcNewIntegerCurrent( LeftCurrent ),
					RightCurrent16Bit	= CalcNewIntegerCurrent( RightCurrent );

	Scanning->Prescanning = true;

	SetYDACStepDirectionTo( UP );

	for( y=0; y <= argNumberOfPrescans && AcquisitationInProgress; y++)
	{
		// Right Scan:
		// The driver global variables, RightBias, LeftBias, RightCurrent, LeftCurrent and set in
		// the ioctl()-call STM_PREPAE_SCAN below.
		// Left and right is refering to the window to which the data will be presented.

		SetIntegerBiasAndCurrent( LeftBias16Bit, LeftCurrent16Bit );
		SetXDACStepDirectionTo( UP );
		
		microsecond_delay( Timing->LineDelay );
	
		Scanning->ScanningDirectionIsRight = true;
	
		for(x=0; x < XSize && AcquisitationInProgress; x++)
		{
			uint16	Sample;
		
			microsecond_delay(					Timing->SettlingTime );
			Sample 								= TakeSamplesFromADC( ImageParams->SamplesPerPixel );
			LeftRawDataBuffer[ x + y * XSize ]	= Sample;
			StepOneXPixel( 						ImageParams->DAStepsPerPixel, Timing->DADelay );
			ValueOfXDAC							+= ImageParams->DAStepsPerPixel;
			Scanning->ScanXPos					= x;
			
		}

		microsecond_delay( Timing->LineDelay );

		// Left Scan:
		// The driver global variables, RightBias, LeftBias, RightCurrent, LeftCurrent and set in
		// the ioctl()-call STM_PREPAE_SCAN below.
		// Left and right is refering to the window to which the data will be presented.
		
		SetIntegerBiasAndCurrent( RightBias16Bit, RightCurrent16Bit );
		SetXDACStepDirectionTo( DOWN );

		Scanning->ScanningDirectionIsRight = false;

		if( AcquisitationInProgress )
			for(x=XSize; x > 0 && AcquisitationInProgress; x--)
			{
				microsecond_delay( 						Timing->SettlingTime );
				RightRawDataBuffer[ x + y * XSize]		= TakeSamplesFromADC( ImageParams->SamplesPerPixel );
				StepOneXPixel(							ImageParams->DAStepsPerPixel, Timing->DADelay );
				ValueOfXDAC								-= ImageParams->DAStepsPerPixel;
				Scanning->ScanXPos 						= x;
			}

		microsecond_delay( Timing->LineDelay );
		Scanning->ScanYPos = y;

	}	

	Scanning->Prescanning = false;
}

int32
RealScan(			void *)
{
	// The variables Timing and ImageParams are set in
	// the ioctl()-call STM_PREPAE_SCAN below.
	uint16			x					= 0,
					y					= 0,
					XSize				= Scanning->ScanXSize,
					YSize				= Scanning->ScanYSize,
	// Left/RightBias/Current variables are set in 
	// the ioctl()-call STM_PREPAE_SCAN below.
					LeftBias16Bit		= CalcNewIntegerBias( LeftBias ),
					RightBias16Bit		= CalcNewIntegerBias( RightBias ),
					LeftCurrent16Bit	= CalcNewIntegerCurrent( LeftCurrent ),
					RightCurrent16Bit	= CalcNewIntegerCurrent( RightCurrent );

	DoPreScans( Timing->NumberOfPrescans );

	MoveTipToDAPoint( 	Scanning->StartXPos, Scanning->StartYPos );

	Timing->StartTime = time( NULL );
	SetYDACStepDirectionTo( UP );

	for( y=0; y <= YSize && AcquisitationInProgress; y++)
	{
		// Right Scan:
		// The driver global variables, RightBias, LeftBias, RightCurrent, LeftCurrent and set in
		// the ioctl()-call STM_PREPAE_SCAN below.
		// Left and right is refering to the window to which the data will be presented.

		SetIntegerBiasAndCurrent( LeftBias16Bit, LeftCurrent16Bit );
		SetXDACStepDirectionTo( UP );
		
		microsecond_delay( Timing->LineDelay );
	
		Scanning->ScanningDirectionIsRight = true;
	
		for(x=0; x < XSize && AcquisitationInProgress; x++)
		{
			uint16	Sample;
		
			microsecond_delay(					Timing->SettlingTime );
			Sample 								= TakeSamplesFromADC( ImageParams->SamplesPerPixel );
			LeftRawDataBuffer[ x + y * XSize ]	= Sample;
			StepOneXPixel( 						ImageParams->DAStepsPerPixel, Timing->DADelay );
			ValueOfXDAC							+= ImageParams->DAStepsPerPixel;
			Scanning->ScanXPos					= x;
			
		}

		microsecond_delay( Timing->LineDelay );

		// Left Scan:
		// The driver global variables, RightBias, LeftBias, RightCurrent, LeftCurrent and set in
		// the ioctl()-call STM_PREPAE_SCAN below.
		// Left and right is refering to the window to which the data will be presented.
		
		SetIntegerBiasAndCurrent( RightBias16Bit, RightCurrent16Bit );
		SetXDACStepDirectionTo( DOWN );

		Scanning->ScanningDirectionIsRight = false;

		if( AcquisitationInProgress )
			for(x=XSize; x > 0 && AcquisitationInProgress; x--)
			{
				microsecond_delay( 						Timing->SettlingTime );
				RightRawDataBuffer[ x + y * XSize]		= TakeSamplesFromADC( ImageParams->SamplesPerPixel );
				StepOneXPixel(							ImageParams->DAStepsPerPixel, Timing->DADelay );
				ValueOfXDAC								-= ImageParams->DAStepsPerPixel;
				Scanning->ScanXPos 						= x;
			}

		StepOneYPixel( ImageParams->DAStepsPerPixel, Timing->DADelay );
		ValueOfYDAC									+= ImageParams->DAStepsPerPixel;
		microsecond_delay( Timing->LineDelay );
		Scanning->ScanYPos = y;

	}	
	Timing->StopTime = time( NULL );

	MoveTipToDAPoint( 	Scanning->StartXPos, Scanning->StartYPos );

	AcquisitationInProgress = false;
	return B_OK;
}

status_t
K575_open( const char *name, uint32 flags, void **cookie )
{
	name;
	flags;
	cookie;
	long		*me;
	
	ENABLE_DEBUG
	D("open()\n");

	if( DriverIsOpen )
		D("Already open, but I'll try to continue. Hope it works.\n");

	// Check if connection with K575 and see if AMM2 module is installed:

	// Initialize the ADC and DACs:
	
	int32	ReturnValue = InitADCAndDAC();
	
	if( ReturnValue != B_OK )
	{
		D("InitADC... etc failed!");
		dprintf("In open(): Error code: %d\n", ReturnValue );
		DriverIsOpen = false;
		return ReturnValue;
	}

	me 				= (long *)malloc(sizeof(me));
	me[0]			= 'K575';
	*cookie 		= me;
	DriverIsOpen	= true;

	return B_OK;
}           

status_t
K575_close( void *cookie )
{
	int32 		ExitValue;
	cookie;
		
	ENABLE_DEBUG
	
	D("close()\n");

	if( ! DriverIsOpen )
	{
		D("Driver not open!!\n");
		return B_ERROR;	
	}

	
	if( AcquisitationThreadID < B_NO_ERROR )
	{
		D("No acquisitation thread!\n");
		return B_ERROR;
	}

	AcquisitationInProgress = false;

	if( AcquisitationThreadID > B_ERROR )
		wait_for_thread( AcquisitationThreadID, &ExitValue );
	
	DriverIsOpen = false;
		
	return B_OK;
}

const char **publish_devices()
{
	D("Was called to publish devices!\n");

	return TheDeviceName;
}

status_t
init_hardware(void)
{
	ENABLE_DEBUG

	D("Was called to init hardware!\n");
	return B_OK;
}

device_hooks *
find_device(const char *name)
{
	ENABLE_DEBUG

	dprintf("Was called to find device '%s'.\n", name);
	if( ! strcmp( name, "K575" ) )
		return &devices[0];
	else
		return NULL;
}

status_t
init_driver(void)
{
	ENABLE_DEBUG

	D("Was called to init driver!\n");

	area_id 	ISAAreaID;
	area_info 	TransferAreaInfo;

	if( ( ISAAreaID = find_area("isa_mem") ) < B_NO_ERROR )
	{
		D("Could not find the ISA memory.\n");
		return B_ERROR;
	}

	
	if ( get_area_info( ISAAreaID, &TransferAreaInfo ) < B_NO_ERROR )
	{
		D("Could not get area info for ISA memory.\n");
		return B_ERROR;
	}

	ISA_Memory = TransferAreaInfo.address;

	return B_OK;
}

void
uninit_driver(void)
{
	ENABLE_DEBUG

	D("Was called to uninit driver!\n");
	
	if( DriverIsOpen )
		K575_close( NULL );
}

static status_t
K575_free_cookie( void *cookie)
{
	ENABLE_DEBUG

	if( cookie )
		free( cookie );

	return B_OK;
}

status_t
K575_control( void *cookie, uint32 op, void *data, size_t len )
{
	cookie;	
	len;
	dprintf("called control().\n");

	// Here is 'data' a pointer to a STMWindow object if the call
	// is a custom call. Unless in the allocation calls, because
	// then it is the size of the area.
	// Otherwise, it depends on that action the caller calls for.
	
	ENABLE_DEBUG

	switch( op )
	{
		case B_GET_DEVICE_SIZE:
			D("called control(). CODE:B_GET_DEVICE_SIZE\n");
			*(long*)data=0;
			break;
		
		case B_SET_DEVICE_SIZE:
			D("called control(). CODE:B_SET_DEVICE_SIZE\n");
			if( (long)data > 0)
				return B_ERROR;
			else
				break; 

		case B_SET_BLOCKING_IO:
			D("called control(). CODE:B_SET_BLOCKING_IO\n");
			break;

		case B_SET_NONBLOCKING_IO:	
			D("called control(). CODE:B_SET_NONBLOCKING_IO\n");
			return B_ERROR;				// The data acquisitation procedure is always blocking!
		
		case B_GET_READ_STATUS:
			D("called control(). CODE:B_GET_READ_STATUS\n");
			*(long*)data = !AcquisitationInProgress;
			break;
			
		case B_GET_WRITE_STATUS:
			D("called control(). CODE:B_GET_WRITE_STATUS\n");
			*(long*)data = false;		// There's no purpose to try to write anything in this driver!
			return B_ERROR;

		case B_GET_GEOMETRY:
			D("called control(). CODE:B_GET_GEOMETRY\n");
			return B_ERROR;				// It has no geometry, because it is not a mass storage unit!

		case B_GET_DRIVER_FOR_DEVICE:
			D("called control(). CODE:B_GET_DRIVER_FOR_DEVICE\n");
			return B_ERROR;

		case B_GET_PARTITION_INFO:
			D("called control(). CODE:B_GET_PARTITION_INFO\n");
			return B_ERROR;

		case B_FORMAT_DEVICE:
			D("called control(). CODE:B_FORMAT_DEVICE\n");
			return B_ERROR;				// Doesn't make any sense.

		case B_EJECT_DEVICE:			// Non ejectable.
			D("called control(). CODE:B_EJECT_DEVICE\n");
			return B_ERROR;

		case B_GET_ICON:
			D("called control(). CODE:B_GET_ICON\n");
			return B_ERROR;	

		//			
		// Below begins the STM-specific controls.
		//

		case STM_PREPARE_SCAN:
			// We MUST call this before performing a Scan. The X and Y positions
			// must be set to zero. The Conversion Thread (in ImageWindow) waits
			// for these positions, and we want it to start at (0,0).)
			// The sequence of events is: 1. Prepare scan. 2. Fire up conversion and
			// plotting threads. (They are waiting for the coordinates to move.)
			// 3. Start the scan below. The rest of the initiation could just as
			// well been done within the starting procedure, but I chose to put it
			// here.	get_area_info( Kernel_RawDataID, &RawDataInfo );
	
			D("called control(). CODE:STM_PREPARE_SCAN\n");
			CallingWindow		= (STMImageWindow *)data;
			Scanning 			= CallingWindow->GetScanningParameters()->Scanning;
			Timing 				= CallingWindow->GetTimingParameters()->Timing;
			ImageParams			= CallingWindow->GetImageParameters()->ImageParams;
			LeftRawDataBuffer	= CallingWindow->GetLeftImageView()->Image->GetKernelRawData();
			dprintf("LeftRawDataBuffer = %x\n",LeftRawDataBuffer );
			RightRawDataBuffer	= CallingWindow->GetRightImageView()->Image->GetKernelRawData();
			dprintf("RightRawDataBuffer = %x\n",RightRawDataBuffer );
			LeftBias			= CallingWindow->GetLeftImageView()->Image->GetBias();
			RightBias			= CallingWindow->GetRightImageView()->Image->GetBias();
			LeftCurrent			= CallingWindow->GetLeftImageView()->Image->GetCurrent();
			RightCurrent		= CallingWindow->GetRightImageView()->Image->GetCurrent();
			MoveTipToDAPoint( 	Scanning->StartXPos, Scanning->StartYPos );
			Scanning->ScanYPos 	= 0;
			Scanning->ScanXPos	= 0;
			ConfigADCAndDACFromParameters( CallingWindow->GetADDAParameters()->ADDA );
			BiasDARange = 10;	// xxx
			SetDARange( 0, BiasDARange );
			CurrentDARange = 10; //xxx
			SetDARange( 1, CurrentDARange );
			break;

		case STM_BEGIN_SCAN:
			// Here we initiate the scanning procedure.
			D("called control(). CODE:STM_BEGIN_SCAN\n");
			AcquisitationInProgress = true;		

			if( data == NULL )
			{
				dprintf("K575:control: data == NULL !\n");
				return B_ERROR;
			}

			AcquisitationThreadID = spawn_kernel_thread(	RealScan, 
															"Real Scan Thread",
															B_REAL_TIME_PRIORITY,
															data);
	
			if( AcquisitationThreadID < B_NO_ERROR )
			{
				D("No acquisitation thread!\n");
				return B_ERROR;
			}
			resume_thread( AcquisitationThreadID );
			break;							
			
		case STM_STOP_SCANNING:					// This stops the procedure of scanning immediately.
			D("called control(). CODE:STM_STOP_SCANNING\n");
			int32	ExitCode;
			AcquisitationInProgress = false;
			wait_for_thread( AcquisitationThreadID, &ExitCode );
			break;
			
		case STM_FINISH_THIS_SCAN:
			D("called control(). CODE:STM_FINISH_THIS_SCAN\n");
												// Here we wait some time before stopping the actual scanning.
			AcquisitationInProgress = false;
			break;
			
		case STM_SEND_PULSES:					// Here we override the voltages set and step in and set the voltages
			D("called control(). CODE:STM_SEND_PULSES\n");
			break;								// according to the specified voltage.
			
		case STM_ALLOCATE_KERNEL_AREA:	
			D("called control(). CODE:STM_ALLOCATE_KERNEL_AREA\n");
			dprintf("data =%d\n", data);
			char			AreaName[30];
			sprintf(AreaName, "Kernel area #d", KernelAreaCounter++);
			void *Kernel_RawData;
					
			return create_area( 	AreaName, 
									&Kernel_RawData, 
									B_ANY_KERNEL_ADDRESS,
									B_PAGE_SIZE*((int32)data/B_PAGE_SIZE+1),
									B_CONTIGUOUS,
									B_READ_AREA | B_WRITE_AREA );
		break;
			
		case STM_DELETE_KERNEL_AREA:
			D("called control(). CODE:STM_DELETE_KERNEL_AREA\n");
			if( (int32)data != B_ERROR )	// Yes, I can check it that way. If error, I set it to B_ERROR. Under abort:.
				delete_area( (int32)data );				
			break;
		
		case STM_SET_BIAS_AND_CURRENT:
			D("called control(). CODE:STM_SET_BIAS_AND_CURRENT\n");
			CallingWindow		= (STMImageWindow *)data;
			Timing 				= CallingWindow->GetTimingParameters()->Timing;
			ConfigADCAndDACFromParameters( CallingWindow->GetADDAParameters()->ADDA );
			BiasDARange = 10;	// xxx
			SetDARange( 0, BiasDARange );
			CurrentDARange = 10; //xxx
			SetDARange( 1, CurrentDARange );
			SetBiasAndCurrent( CallingWindow->GetLeftImageView()->Image->GetBias(), CallingWindow->GetLeftImageView()->Image->GetCurrent() );
			break;
		
		default:								// This is an unrecognized command.
			dprintf("called control(). CODE:default: %d\n", op);
			return B_ERROR;
	}

	return B_NO_ERROR;
} 

status_t
K575_read( void *cookie, off_t position, void *data, size_t *numBytes )
{
	ENABLE_DEBUG

	D("called read().\n");

	cookie;
	position;
	data;
	numBytes;
	
	return B_NO_ERROR;
} 



status_t
K575_write(void *cookie, off_t position,const void *data, size_t *numBytes)
{
	ENABLE_DEBUG

	D("called write().\n");
	
	cookie;
	position;
	data;
	numBytes;

	return B_NO_ERROR;
}
